"""This module contains a function :func:`compute_sun_position` to 
calculate the sun position for a given time and location.

"""
from math import asin, atan, cos, degrees, radians, sin, tan


def compute_sun_position(time, lat, lon):
    """Compute the sun position.

    Computes the sun position for a given time and geographical location.
    Detailed calculation is explained in Wikipedia article Sonnenstand_.
    For calculation of right ascension alpha and declination delta the article
    refers to [#]_. For calculation of eclipitc longitude lambda the aricle
    refers to [#]_.

    Parameters
    ----------
    time : int
        Time as unix timestamp.
    lat : float
        Geographical latitude in degree (+North, -South).
    lon : float
        Geographical longitude in degree (+East, -West).

    Returns
    -------
    dict
        *dict* with keys *'azimuth'* and *'elevation'*.

        - *'azimuth'* - Solar azimuth angle in degree (Noth 0°, East 90°,
          West 270°)
        - *'elevation'* - Elevation of the sun in degree

    .. _Sonnenstand: https://de.wikipedia.org/wiki/Sonnenstand

    .. [#] The Astronomical Almanac For The Year 2006, The Stationary Office,
           London 2004

    .. [#] Astronomical Algorithms, Billmann-Bell, Richmond 2000 (2nd ed.,
           2nd printing)

    """

    JD_1970 = 2440587.5  # julian day of epoch 1970-01-01 00:00
    JD_time = time / 86400.0 + JD_1970
    # print(JD_time)
    JD_2000 = 2451545.0  # Greenich noon 2000-01-01 12:00
    n = JD_time - JD_2000  # days since Greenich noon 2000-01-01

    # Calculate the mean longitude of the Sun, corrected for the aberration of
    # light
    L = (280.460 + 0.9856474 * n) % 360  # degree

    # Calculate the mean anomaly of the sun
    g = (357.528 + 0.9856003 * n) % 360  # degree

    # Calculate the ecliptic longitude of the sun
    Lambda = L + 1.915 * sin(radians(g)) + 0.01997 * sin(radians(2 * g))

    # Calculate obliquity of the ecliptic
    epsilon = 23.439 - 4e-7 * n

    # Calculate right ascension
    denominator = cos(radians(Lambda))
    alpha = degrees(
        atan(cos(radians(epsilon)) * sin(radians(Lambda)) / denominator)
    )

    # Bring angle to apropriate quadrant
    # print(denominator)
    if denominator < 0:
        alpha = alpha + 180

    # Calculate declination
    delta = degrees(asin(sin(radians(epsilon)) * sin(radians(Lambda))))
    # print(alpha, delta)

    # Calculate Greenwich hour angle in several steps

    # First calculate julian day of time 00:00 UTC
    JD_0 = time // 86400 + JD_1970

    # Calculate julian centuries since UTC 2000-01-01 00:00
    T_0 = (JD_0 - JD_2000) / 36525
    # print("T_0: %s" % T_0)

    # Calculate hour of day, for example 17:45 -> 17.75
    T = (time % 86400) / 3600

    # Calculate Greenwich mean sidereal time
    teta_G_h = (6.697376 + 2400.05134 * T_0 + 1.002738 * T) % 24
    # First term describes Greenwich sidereal time at 2000-01-01 00:00.
    # Second term describes the daily advancement of sidereal time for about 4
    # minutes per day.
    # The third term adds the fraction of the day meassured in sidereal time

    # Convert from hours to degree (1h = 15°)
    teta_G = teta_G_h * 15

    # Consider geographical longitude (count eastern direction positive)
    teta = teta_G + lon
    # print("teta_G: %s" % teta_G)

    # Calculate hour angle for this place by subtracting right ascension
    tau = teta - alpha

    # Calculate azimuth
    denominator = cos(radians(tau)) * sin(radians(lat)) - tan(
        radians(delta)
    ) * cos(radians(lat))
    azimuth = degrees(atan(sin(radians(tau)) / denominator))

    # Bring azimuth to appropriate quadrant
    if denominator < 0:
        azimuth = azimuth + 180

    # Count azimuth from north
    azimuth = (azimuth + 180) % 360

    # Calculate elevation angle of sun
    h = degrees(
        asin(
            cos(radians(delta)) * cos(radians(tau)) * cos(radians(lat))
            + sin(radians(delta)) * sin(radians(lat))
        )
    )
    # print("Azimuth: %s, h: %s" % (azimuth, h))

    # Coorect elevation because of atmospheric refraction
    R = 1.02 / tan(radians(h + 10.3 / (h + 5.11)))
    h_R = h + R / 60
    # Keep in mind that refraction depends on detailed state of atmosphere.
    # The given formula assumes air pressure 1010 mbar and temperatur 10°C

    return {"azimuth": azimuth, "elevation": h_R}



"""This module contains the flexibility model."""
from datetime import datetime, timedelta
from typing import Optional

import numpy as np
import pandas as pd
from pysimmods.util.date_util import GER, TZ

from . import LOG
from .forecast_model import ForecastModel


class FlexibilityModel(ForecastModel):
    """The flexibility model for all pysimmods."""

    def __init__(
        self,
        model,
        unit="kw",
        priorize_setpoint=False,
        step_size: Optional[int] = None,
        now_dt: Optional[datetime] = None,
        forecast_horizon_hours=1,
        seed=None,
    ):
        super().__init__(
            model,
            unit,
            priorize_setpoint,
            step_size,
            now_dt,
            forecast_horizon_hours,
        )

        if seed is not None:
            self._rng = np.random.RandomState(seed)
        else:
            self._rng = np.random.RandomState()

    def generate_schedules(
        self, start, flexibility_horizon_hours, num_schedules
    ):
        """Perform sampling and generate a set of schedules for the
        specified time interval.

        Args:
            start (str): Is the start of the planning horizon for which
                the sampling is done. It has to be provided as ISO 8601
                timezone string such as '2020-06-22 12:00:00+0000'

            forecast_horizon_hours (int): The planning horizon is divided into.

        """
        state_backup = self._model.get_state()
        step_size = self._step_size

        now_dt = self._now_dt
        start_dt = datetime.strptime(start, GER)
        end_dt = (
            start_dt
            + timedelta(hours=flexibility_horizon_hours)
            - timedelta(seconds=step_size)
        )
        periods = flexibility_horizon_hours * 3_600 / step_size

        # Fast forward to the planning interval
        while start_dt > now_dt:
            try:
                self._calculate_step(
                    now_dt,
                    self.schedule.get(now_dt, "target") / self._percent_factor,
                )
                now_dt += timedelta(seconds=step_size)
            except KeyError as err:
                # raise err
                LOG.info("Could not create flexibilities: %s", err)
                return dict()

        index = pd.date_range(start_dt, end_dt, periods=periods)

        self.flexibilities = dict()

        for schedule_id in range(num_schedules):
            schedule = self._generate_schedule(
                pd.DataFrame(
                    columns=["target", self._pname, self._qname],
                    index=index,
                )
            )

            self.flexibilities[schedule_id] = schedule.tz_convert(tz=TZ)

        self._model.set_state(state_backup)
        return self.flexibilities

    def _generate_schedule(self, dataframe):
        num_steps = len(dataframe.index)
        dataframe["target"] = self._rng.uniform(size=num_steps) / (
            0.01 / self._percent_factor
        )

        state_backup = self._model.get_state()
        for index, row in dataframe.iterrows():
            try:
                self._calculate_step(index, row["target"])
                dataframe.loc[index, self._pname] = (
                    self._model.get_p_kw() * self._unit_factor
                )
                dataframe.loc[index, self._qname] = (
                    self._model.get_q_kvar() * self._unit_factor
                )

            except KeyError:
                # Forecast is missing
                dataframe.loc[index, self._pname] = np.nan
                dataframe.loc[index, self._qname] = np.nan
                dataframe.loc[index, "target"] = np.nan
        self._model.set_state(state_backup)

        return dataframe



from datetime import datetime, timezone

NOTZ = "%Y-%m-%d %H:%M:%S"
GER = "%Y-%m-%d %H:%M:%S%z"
TZ = "Europe/Berlin"


def convert_dt(now):
    if isinstance(now, str):
        now = datetime.strptime(now, GER)
    elif isinstance(now, int):
        now = datetime.utcfromtimestamp(now)

    if isinstance(now, datetime):
        now = now.astimezone(timezone.utc)
    else:
        raise ValueError(f"Unknown date format: {now} of type {type(now)}")

    return now



"""This module contains multiple configuration examples for the
biogas plant.

"""

import sys

ALL_PRESETS = [
    (40, 1),
    (80, 1),
    (320, 1),
    (550, 1),
    (373, 1),
    (1250, 1),
    (55, 2),
    (110, 2),
    (220, 2),
    (300, 2),
    (320, 2),
    (2050, 2),
    (555, 3),
    (1500, 3),
]


def biogas_preset(pn_max_kw, **kwargs):
    """Return the parameter configuration for a biogas model from the
    pysimmods package.

    """
    thismodule = sys.modules[__name__]
    num_gens = kwargs.get("num_gens", None)

    possible_p = [val for val in ALL_PRESETS if val[0] == pn_max_kw]

    if num_gens is not None:
        possible_p = [val for val in possible_p if val[1] == num_gens]

    method_params = f"params_{possible_p[0][1]}g_{possible_p[0][0]}kw"
    method_inits = f"inits_{possible_p[0][1]}g_{possible_p[0][0]}kw"

    params = getattr(thismodule, method_params)()
    params["sign_convention"] = "active"

    return params, getattr(thismodule, method_inits)()


def params_1g_40kw():
    """Params for a Biogas plant with 40 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 550,
        "cap_gas_m3": 250,
        "gas_fill_min_percent": 2.5,
        "gas_fill_max_percent": 97.5,
        "ch4_concentration_percent": 51.9,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [20, 40],
            "eta_stages_percent": [28.0, 30.0],
            "eta_th_stages_percent": [31.4, 32.0],
            "restarts_per_day": 0,
            "active_min_s": 0,
            "active_max_s_per_day": 18 * 3_600,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_40kw():
    """Init vals for a Biogas plant with 40 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 20,
        },
    }


def params_1g_80kw():
    """Params for a Biogas plant with 80 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 750,
        "cap_gas_m3": 375,
        "gas_fill_min_percent": 2.5,
        "gas_fill_max_percent": 97.5,
        "ch4_concentration_percent": 52.9,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [40, 80],
            "eta_stages_percent": [37.0, 36.7],
            "eta_th_stages_percent": [32.4, 32.0],
            "restarts_per_day": 5,
            "active_min_s": 0,
            "active_max_s_per_day": 18 * 3_600,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_80kw():
    """Init vals for a Biogas plant with 80 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 40,
        },
    }


def params_1g_320kw():
    """Params for a Biogas plant with 320 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 2100,
        "cap_gas_m3": 2000,
        "gas_fill_min_percent": 5.0,
        "gas_fill_max_percent": 95.0,
        "ch4_concentration_percent": 55.3,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [160, 240, 320],
            "eta_stages_percent": [36.6, 37.0, 37.5],
            "eta_th_stages_percent": [31.0, 31.5, 31.9],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_320kw():
    """Init vals for a Biogas plant with 320 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 160,
        },
    }


def params_1g_550kw():
    """Params for a Biogas plant with 550 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 5450,
        "cap_gas_m3": 1578,
        "gas_fill_min_percent": 2.0,
        "gas_fill_max_percent": 98.0,
        "ch4_concentration_percent": 55.4,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [275, 330, 440, 550],
            "eta_stages_percent": [31.6, 31.8, 32.0, 32.2],
            "eta_th_stages_percent": [41.0, 42.1, 43.2, 44.2],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_550kw():
    """Init vals for a Biogas plant with 550 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 275,
        },
    }


def params_1g_373kw():
    """Params for a Biogas plant with 375 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 2550,
        "cap_gas_m3": 2500,
        "gas_fill_min_percent": 5.0,
        "gas_fill_max_percent": 95.0,
        "ch4_concentration_percent": 58.5,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [121, 252, 373],
            "eta_stages_percent": [33.7, 36.2, 37.9],
            "eta_th_stages_percent": [45.0, 46.0, 47.8],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_373kw():
    """Init vals for a Biogas plant with 373 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 750,
        },
    }


def params_1g_1250kw():
    """Params for a Biogas plant with 1250 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_m3_per_day": 9_750,
        "cap_gas_m3": 9_250,
        "gas_fill_min_percent": 2.5,
        "gas_fill_max_percent": 97.5,
        "ch4_concentration_percent": 50.0,
        "num_chps": 1,
        "chp0": {
            "pn_stages_kw": [750, 1250],
            "eta_stages_percent": [37.0, 41.25],
            "eta_th_stages_percent": [48.0, 44.0],
            "restarts_per_day": 4,
            "active_min_s": 0,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_1g_1250kw():
    """Init vals for a Biogas plant with 1250 kw nominal power and
    one CHP CNG.
    """
    return {
        "gas_fill_percent": 50.0,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 750,
        },
    }


def params_2g_55kw():
    """Params for a Biogas plant with 55 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 255,
        "cap_gas_m3": 390,
        "gas_fill_min_percent": 5.0,
        "gas_fill_max_percent": 95.0,
        "ch4_concentration_percent": 50.2,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [9.0, 18.0],
            "eta_stages_percent": [29.3, 28.7],
            "eta_th_stages_percent": [35.4, 36.1],
            "restarts_per_day": 0,
            "active_min_s": 1 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [18.5, 37.0],
            "eta_stages_percent": [31.4, 32.5],
            "eta_th_stages_percent": [36.1, 36.8],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_55kw():
    """Init vals for a Biogas plant with 55 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 27.5,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 18.3,
        },
    }


def params_2g_110kw():
    """Params for a Biogas plant with 110 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 950,
        "cap_gas_m3": 1000,
        "gas_fill_min_percent": 2.5,
        "gas_fill_max_percent": 97.5,
        "ch4_concentration_percent": 56.8,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [27.5, 55.0],
            "eta_stages_percent": [29.7, 31.9],
            "eta_th_stages_percent": [36.4, 37.0],
            "restarts_per_day": 6,
            "active_min_s": 1 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [18.3, 36.6, 55.0],
            "eta_stages_percent": [30.8, 32.2, 33.9],
            "eta_th_stages_percent": [36.1, 36.8, 38.0],
            "restarts_per_day": 6,
            "active_min_s": 1 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_110kw():
    """Init vals for a Biogas plant with 110 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 27.5,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 18.3,
        },
    }


def params_2g_220kw():
    """Params for a Biogas plant with 220 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 2000,
        "cap_gas_m3": 1150,
        "gas_fill_min_percent": 2.5,
        "gas_fill_max_percent": 97.5,
        "ch4_concentration_percent": 55.7,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [55.0, 110.0],
            "eta_stages_percent": [35.4, 34.7],
            "eta_th_stages_percent": [31.4, 31.0],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [55.0, 110.0],
            "eta_stages_percent": [35.5, 36.1],
            "eta_th_stages_percent": [38.0, 42.4],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_220kw():
    """Init vals for a Biogas plant with 220 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 55.0,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 55.0,
        },
    }


def params_2g_300kw():
    """Params for a Biogas plant with 300 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 3030,
        "cap_gas_m3": 1530,
        "gas_fill_min_percent": 2.0,
        "gas_fill_max_percent": 98.0,
        "ch4_concentration_percent": 54.4,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [75.0, 150.0],
            "eta_stages_percent": [31.1, 32.4],
            "eta_th_stages_percent": [30.8, 31.2],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 2 * 3_600,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [75.0, 150.0],
            "eta_stages_percent": [33.0, 33.2],
            "eta_th_stages_percent": [30.0, 30.5],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 2 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_300kw():
    """Init vals for a Biogas plant with 300 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 75.0,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 75.0,
        },
    }


def params_2g_320kw():
    """Params for a Biogas plant with 320 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 2050,
        "cap_gas_m3": 855,
        "gas_fill_min_percent": 2.0,
        "gas_fill_max_percent": 98.0,
        "ch4_concentration_percent": 63.9,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [40.0, 80.0, 120.0, 160.0],
            "eta_stages_percent": [31.1, 32.2, 33.3, 34.4],
            "eta_th_stages_percent": [30.8, 31.2, 32.3, 32.3],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [85.0, 110.0, 135.0, 160.0],
            "eta_stages_percent": [32.2, 33.0, 33.8, 34.6],
            "eta_th_stages_percent": [30.0, 30.5, 31.0, 31.5],
            "restarts_per_day": 6,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_320kw():
    """Init vals for a Biogas plant with 320 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 80.0,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 85.0,
        },
    }


def params_2g_2050kw():
    """Params for a Biogas plant with 2050 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_m3_per_day": 14_500,
        "cap_gas_m3": 13_000,
        "gas_fill_min_percent": 5,
        "gas_fill_max_percent": 95,
        "ch4_concentration_percent": 50.0,
        "has_heat_storage": False,
        "num_chps": 2,
        "chp0": {
            "pn_stages_kw": [700, 1150],
            "eta_stages_percent": [37.0, 44.0],
            "eta_th_stages_percent": [40.0, 44.0],
            "restarts_per_day": 4,
            "active_min_s": 0,
            "active_max_s_per_day": 16 * 3_600,
            "inactive_min_s": 2 * 3_600,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [600, 900],
            "eta_stages_percent": [37, 40.0],
            "eta_th_stages_percent": [42.0, 43.0],
            "restarts_per_day": 4,
            "active_min_s": 0,
            "active_max_s_per_day": 16 * 3_600,
            "inactive_min_s": 2 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_2g_2050kw():
    """Init vals for a Biogas plant with 2050 kw nominal power and
    two CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 700,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 600,
        },
    }


def params_3g_555kw():
    """Params for a Biogas plant with 555 kw nominal power and
    three CHP CNGs.
    """
    return {
        "gas_m3_per_day": 6000,
        "cap_gas_m3": 2500,
        "gas_fill_min_percent": 5,
        "gas_fill_max_percent": 95,
        "ch4_concentration_percent": 51.6,
        "has_heat_storage": False,
        "num_chps": 3,
        "chp0": {
            "pn_stages_kw": [125, 250],
            "eta_stages_percent": [31.2, 34.7],
            "eta_th_stages_percent": [42.0, 45.9],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 18 * 3_600,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [125, 250],
            "eta_stages_percent": [34.3, 36.1],
            "eta_th_stages_percent": [41.9, 43],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 18 * 3_600,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
        "chp2": {
            "pn_stages_kw": [27.5, 55],
            "eta_stages_percent": [28.6, 30],
            "eta_th_stages_percent": [35.1, 36.1],
            "restarts_per_day": 0,
            "active_min_s": 2 * 3_600,
            "active_max_s_per_day": 18 * 3_600,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_3g_555kw():
    """Init vals for a Biogas plant with 555 kw nominal power and
    three CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 250,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 125,
        },
        "chp2": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 55,
        },
    }


def params_3g_1500kw():
    """Params for a Biogas plant with 1500 kw nominal power and
    three CHP CNGs.
    """
    return {
        "gas_m3_per_day": 16_040,
        "cap_gas_m3": 25_200,
        "gas_fill_min_percent": 5,
        "gas_fill_max_percent": 95,
        "ch4_concentration_percent": 53.9,
        "has_heat_storage": False,
        "num_chps": 3,
        "chp0": {
            "pn_stages_kw": [
                125.0,
                250.0,
                375.0,
                500.0,
                625.0,
                750.0,
            ],
            "eta_stages_percent": [
                26.1,
                27.0,
                27.3,
                28.0,
                27.7,
                27.5,
            ],
            "eta_th_stages_percent": [
                50.0,
                50.2,
                50.5,
                50.9,
                51.0,
                52.1,
            ],
            "restarts_per_day": 0,
            "active_min_s": 3 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
        "chp1": {
            "pn_stages_kw": [330.0, 440.0, 550],
            "eta_stages_percent": [27.9, 28.5, 29.1],
            "eta_th_stages_percent": [45.9, 47.3, 51.0],
            "restarts_per_day": 0,
            "active_min_s": 1 * 3_600,
            "active_max_s_per_day": 0,
            "inactive_min_s": 0,
            "inactive_max_s_per_day": 0,
        },
        "chp2": {
            "pn_stages_kw": [100, 200],
            "eta_stages_percent": [28.8, 29.9],
            "eta_th_stages_percent": [29.7, 30.1],
            "restarts_per_day": 0,
            "active_min_s": 0,
            "active_max_s_per_day": 0,
            "inactive_min_s": 1 * 3_600,
            "inactive_max_s_per_day": 0,
        },
    }


def inits_3g_1500kw():
    """Init vals for a Biogas plant with 1500 kw nominal power and
    three CHP CNGs.
    """
    return {
        "gas_fill_percent": 50,
        "chp0": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 125.0,
        },
        "chp1": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 330,
        },
        "chp2": {
            "active_s": 0,
            "active_s_per_day": 0,
            "inactive_s": 0,
            "inactive_s_per_day": 0,
            "restarts": 0,
            "p_kw": 100,
        },
    }



"""This module contains the input model for PV plant."""
from pysimmods.model.inputs import ModelInputs


class PVInputs(ModelInputs):
    """Input variables of PV plant model.

    See :class:`pysimmods.model.inputs.ModelInputs` for additional
    information. This class has no inputs itself. Instead, each
    of the values is to be provided before each step.

    Attributes
    ----------
    bh_w_per_m2 : float
        Beam horizontal (direct solar radiation on horizontal plane)
        in [W/m²].
    dh_w_per_m2 : float
        Diffuse horizontal (diffuse solar radiation on horizontal
        plane) in [W/m²].
    s_module_w_per_m2 : float
        Solar irradiance on module surface in [W/m²]. Must be set
        as input instead of :attr:`bh_w_per_m2` and
        :attr:`dh_w_per_m2`, if
        :attr:`~.PVConfig.has_external_irradiance_model` is set to
        True. Otherwise, this attribute is ignored.
    t_air_deg_celsius:
        Air temperature in [°C].
    """

    def __init__(self):
        super().__init__()
        self.bh_w_per_m2 = None
        self.dh_w_per_m2 = None
        self.s_module_w_per_m2 = None
        self.t_air_deg_celsius = None



"""This module contains the input model for the chp."""
from pysimmods.model.inputs import ModelInputs


class CHPCNGInputs(ModelInputs):
    """A CHP CNG input class."""

    def __init__(self):
        super().__init__()

        self.gas_in_m3 = None
        """Available gas (e.g from a gas storage) in [m^3]."""

        self.gas_critical = False
        """If set to True, this unit will reduce generation."""

        self.heat_critical = False
        """If set to True, this unit will increase generation.
        Has lower priority than gas_critical.
        """

    def reset(self):
        """To be called at the end of each step."""
        for attr in self.__dict__.keys():
            setattr(self, attr, None)

        self.gas_critical = False
        self.heat_critical = False



"""This module contains the config model for the battery model."""
from typing import Any, Dict, List

from ...model.config import ModelConfig

DEFAULT_SCHEDULE: List[float] = (
    [50, 59.451814, 60, 60, 60, 60, 50, 50]
    + [50, 35, 35, 35, 35, 50, 50, 70]
    + [70, 50, 35, 35, 50, 51, 51, 54]
)


class BatteryConfig(ModelConfig):
    """Captures the configuration parameters of the battery model

    On intitialization a dictionay with values for all configuration
    parameters has to be passed.

    The configuration parameters are constant during the simulation
    process. That is they are not manipulated in the step-method of
    the battery model and should not be changed during simulation
    from outside.

    Attributes
    ----------
    cap_kwh : float
        Capacity of the battery in [kWh].
    p_charge_max_kw : float
        Maximum charging (consumption) power of battery in [kW].
    p_discharge_max_kw : float
        Maximum discharging (generation) power of battery in [kW].
    soc_min_percent : float
        Minimum state of charge of battery in [%] of capacity.
    eta_pc : list
        Polynomial coefficients for calculating set power dependent eta.
    """

    def __init__(self, params: Dict[str, Any]):
        super().__init__(params)

        self.cap_kwh: float = abs(params["cap_kwh"])
        self.p_charge_max_kw: float = abs(params["p_charge_max_kw"])
        self.p_charge_min_kw: float = 0
        self.p_discharge_max_kw: float = abs(params["p_discharge_max_kw"])
        self.p_discharge_min_kw: float = 0
        self.soc_min_percent: float = params.get("soc_min_percent", 0.0)
        self.eta_pc: List[float] = params.get(
            "eta_pc", [-2.109566, 0.403556, 97.110770]
        )

        self.default_p_schedule: List[float] = []
        for val in params.get("default_p_schedule", DEFAULT_SCHEDULE):
            val = val * 2 - 100
            if val == 0:
                self.default_p_schedule.append(0.0)
            elif -100 <= val < 0:
                self.default_p_schedule.append(
                    val / 100.0 * self.p_discharge_max_kw
                )
            elif 0 < val <= 100:
                self.default_p_schedule.append(
                    val / 100.0 * self.p_charge_max_kw
                )

        self.default_q_schedule: List[float] = [0.0] * 24



"""This module containts functions to create configurations for the
:class:`.PVPlantSystem` model.

"""


def pv_preset(
    p_peak_kw,
    eta=0.25,
    cos_phi=0.9,
    q_control="prioritize_p",
    inverter_mode="inductive",
):
    """Return a preset configuration for a PV plant.

    Creates configuration for parameters and initialization values.

    Parameters
    ----------
    p_peak_kw : float
        The targeted peak power of the PV plant.
    eta : float, optional
        The efficiency of the module. Default is 0.25 (25 %).
    cos_phi : float, optional
        The cosinus of the phase angle for apparent power calculation.
        Default is 0.9.
    q_control : str, optional
        Define how reactive power is controlled by the inverter. See
        :class:`.Inverter` for more information. Default is *"p_set"*.
    inverter_mode : str, optional
        Define the inverter mode. Can be *"inductive"* or
        *"capactive"*. Default is *"inductive"*.

    Returns
    -------
    tuple
        A *tuple* of two *dicts* with all necessary parameters and
        initialization values for a :class:`.PVPlantSystem`.

    """
    params = {
        "pv": {"a_m2": p_peak_kw / eta, "eta_percent": eta * 100},
        "inverter": {
            "sn_kva": p_peak_kw / cos_phi,
            "q_control": q_control,
            "cos_phi": cos_phi,
            "inverter_mode": inverter_mode,
        },
        "sign_convention": "active",
    }
    inits = {"pv": {"t_module_deg_celsius": 5.0}, "inverter": None}

    return params, inits



"""This module contains an inverter model"""

from copy import copy
from math import sqrt

from pysimmods.model.qgenerator import QControl, QGenerator
from pysimmods.other.invertersim.config import InverterConfig
from pysimmods.other.invertersim.inputs import InverterInputs
from pysimmods.other.invertersim.state import InverterState


class Inverter(QGenerator):
    """Inverter model

    Supports five different modes:

    - *p_set*: p_set_kw is provided, cos phi remains constant and
      q_kvar is calculated with sqrt(s^2-p^2).
    - *q_set*: q_set_kvar is provided, cos phi remains constant and
      p_kw is calculated with sqrt(s^2-q^2)
    - *cos_phi_set*: cos phi is provided, p_kw is calculated from
      p_in_kw and q_kvar is calculated accordingly.
    - *pq_set*: p_set_kw and q_set_kvar are provided, with p being
      prioritized, i.e., q is limited by sqrt(s^2-p^2), and cos phi is
    calculated with p/s.
    - *qp_set*: q_set_kvar and p_set_kw are provided, with q being
      prioritized, i.e., p is limited by sqrt(s^2-q^2), and cos phi is
      calculated with p/s

    Parameters
    ----------
    params : dict
        A *dict* containing configuration parameters.
    inits : dict, optional
        A *dict* containing initialization parameters.

    Attributes
    ----------
    config : :class:`~.InverterConfig`
        Configuration parameters of the inverter.
    state : :class:`~.InverterState`
        Initialization parameters of the inverter.
    inputs : :class:`~.InverterInputs`
        Input parameters of the inverter.

    """

    def __init__(self, params, inits=None):
        if inits is None:
            inits = dict()
        self.config = InverterConfig(params)
        self.state = InverterState(inits)
        self.inputs = InverterInputs()

    def step(self):
        """Perform simulation step"""
        next_state = copy(self.state)

        self._check_inputs()

        next_state._inductive = self.inputs._inductive

        self._calculate(next_state)

        # if self.config.q_control == QControl.P_SET:
        #     self._calc_p_set(next_state)

        # elif self.config.q_control == QControl.Q_SET:
        #     self._calc_q_set(next_state)

        # elif self.config.q_control == QControl.COS_PHI_SET:
        #     self._calc_cos_phi_set(next_state)

        # elif self.config.q_control == QControl.PQ_SET:
        #     self._calc_pq_set(next_state)

        # elif self.config.q_control == QControl.QP_SET:
        #     self._calc_qp_set(next_state)

        # if next_state.inductive:
        #     next_state.q_kvar *= -1

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self):
        if self.inputs.p_set_kw is None:
            self.inputs.p_set_kw = self.inputs.p_in_kw

        # if self.inputs.q_set_kvar is None:
        #     self.inputs.q_set_kvar = self.config.s_max_kva

        if self.inputs.cos_phi_set is None:
            self.inputs.cos_phi_set = self.config.cos_phi

        if self.inputs._inductive is None:
            self.inputs._inductive = self.config.inverter_mode == "inductive"

    def _calculate(self, next_state: InverterState):
        # Can't output more power than is available
        p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

        # Check constraints for apparent power
        s_kva = p_kw / self.inputs.cos_phi_set
        s_kva = min(s_kva, self.config.s_max_kva)

        # Use updated apparent power to calculate active power
        p_max_kw = s_kva * self.inputs.cos_phi_set
        p_kw = min(p_kw, p_max_kw)
        q_max_kvar = sqrt(self.config.s_max_kva**2 - p_kw**2)

        if self.inputs.q_set_kvar is None:
            q_kvar = sqrt(s_kva**2 - p_kw**2)
            q_kvar = min(q_max_kvar, q_kvar)
            if next_state._inductive:
                q_kvar *= -1
        else:
            q_kvar = min(self.inputs.q_set_kvar, self.config.s_max_kva)
            s_kva = sqrt(p_kw**2 + q_kvar**2)

            if s_kva > self.config.s_max_kva:
                if self.config.q_control == QControl.PRIORITIZE_P:
                    q_kvar = q_max_kvar
                elif self.config.q_control == QControl.PRIORITIZE_Q:
                    p_kw = sqrt(self.config.s_max_kva**2 - q_kvar**2)

        if s_kva > 0:
            next_state.cos_phi = p_kw / s_kva
        else:
            next_state.cos_phi = self.inputs.cos_phi_set
        next_state.p_kw = p_kw
        next_state.q_kvar = q_kvar

    # def _calc_p_set(self, next_state):

    #     # Current active power
    #     p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

    #     # Check constraints for apparent power
    #     s_kva = p_kw / self.config.cos_phi
    #     s_kva = min(s_kva, self.config.s_max_kva)

    #     # Check constraints for active power
    #     p_max_kw = s_kva * self.config.cos_phi
    #     p_kw = min(p_kw, p_max_kw)

    #     # Current reactive power
    #     q_kvar = (s_kva**2 - p_kw**2) ** 0.5
    #     q_max_kvar = (self.config.s_max_kva**2 - p_kw**2) ** 0.5
    #     q_kvar = min(q_max_kvar, q_kvar)

    #     # Update state
    #     next_state.p_kw = p_kw
    #     next_state.q_kvar = q_kvar
    #     next_state.cos_phi = self.config.cos_phi

    # def _calc_q_set(self, next_state):

    #     # Check constraints for reactive power
    #     phi = acos(self.config.cos_phi)
    #     q_max_kvar = self.config.s_max_kva * sin(phi)
    #     q_kvar = min(abs(self.inputs.q_set_kvar), q_max_kvar)

    #     # Check constraints for active power
    #     p_max_kw = (self.config.s_max_kva**2 - q_kvar**2) ** 0.5
    #     p_kw = min(abs(self.inputs.p_in_kw), p_max_kw)

    #     # Update state
    #     next_state.q_kvar = q_kvar
    #     next_state.p_kw = p_kw
    #     next_state.cos_phi = self.config.cos_phi

    # def _calc_cos_phi_set(self, next_state):

    #     # Check cos phi input

    #     # Check constraints for apparent power
    #     s_kva = abs(self.inputs.p_in_kw) / self.inputs.cos_phi_set
    #     s_kva = min(s_kva, self.config.s_max_kva)

    #     # Check constraints for active power
    #     p_max_kw = s_kva * self.inputs.cos_phi_set
    #     p_kw = min(abs(self.inputs.p_in_kw), p_max_kw)

    #     # Calculate reactive power
    #     q_kvar = (s_kva**2 - p_kw**2) ** 0.5

    #     # Update state
    #     next_state.p_kw = p_kw
    #     next_state.q_kvar = q_kvar
    #     next_state.cos_phi = self.inputs.cos_phi_set

    # def _calc_pq_set(self, next_state):

    #     # Calculate active power
    #     p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

    #     # Check constraints for reactive power
    #     q_max_kvar = (self.config.s_max_kva**2 - p_kw**2) ** 0.5
    #     q_kvar = min(q_max_kvar, abs(self.inputs.q_set_kvar))

    #     # Calculate cos phi
    #     cos_phi = p_kw / self.config.s_max_kva

    #     # Update state
    #     next_state.p_kw = p_kw
    #     next_state.q_kvar = q_kvar
    #     next_state.cos_phi = cos_phi

    # def _calc_qp_set(self, next_state):

    #     # Calculate reactive power
    #     q_kvar = abs(self.inputs.q_set_kvar)
    #     q_kvar = min(q_kvar, self.config.s_max_kva)

    #     # Calculate active power
    #     p_kw = min(abs(self.inputs.p_in_kw), abs(self.inputs.p_set_kw))

    #     # Check constraints for active power
    #     p_max_kw = (
    #         self.config.s_max_kva**2 - self.inputs.q_set_kvar**2
    #     ) ** 0.5
    #     p_kw = min(p_max_kw, p_kw)

    #     # Calculate cos phi
    #     cos_phi = p_kw / self.config.s_max_kva

    #     # Update state
    #     next_state.p_kw = p_kw
    #     next_state.q_kvar = q_kvar
    #     next_state.cos_phi = cos_phi

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar
        # q_min = self.get_qn_min_kvar()
        # q_max = self.get_qn_max_kvar()
        # if self.config.inverter_mode == "capacitive":
        #     q_kvar *= -1

        # if q_kvar * self.config.gsign > 0:
        #     # Consider capacitive and inductive
        #     q_kvar = abs(q_kvar)
        #     if q_kvar < q_min:
        #         self.inputs.q_set_kvar = 0
        #     else:
        #         self.inputs.q_set_kvar = min(q_max, q_kvar)
        # else:
        #     self.inputs.q_set_kvar = 0



from ...model.inputs import ModelInputs


class BatteryInputs(ModelInputs):
    """Captures the input variables of the battery model

    Input variables can be modified from outside the battery model before
    calling the step-method. They are used as an input for the calculation of
    new values for the state variables in the step-method.

    """



"""This module contains a :class:`mosaik_api.Simulator` for all models
of the pysimmods package.

"""
import logging
import pprint
from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import mosaik_api
import numpy as np
from midas.util.dict_util import strtobool
from midas.util.logging import set_and_init_logger
from midas.util.runtime_config import RuntimeConfig
from pysimmods.model.model import Model
from pysimmods.mosaik import LOG
from pysimmods.util.date_util import GER

from .meta import META, MODELS

VLOG = logging.getLogger("pysimmods.mosaik.verbose")


class PysimmodsSimulator(mosaik_api.Simulator):
    """The Pysimmods simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str
        self.models: Dict[str, Model] = {}
        self.num_models: Dict[str, int] = {}

        self.step_size: int
        self.now_dt: datetime
        self.percent_factor: float = 1.0
        self.key_value_logs: bool

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                "Using default step size of 900."
            )
        self.step_size = sim_params.get("step_size", 900)
        self.now_dt = datetime.strptime(
            sim_params["start_date"], GER
        ).astimezone(timezone.utc)

        use_decimal_percent = sim_params.get("use_decimal_percent", False)
        if not isinstance(use_decimal_percent, bool):
            try:
                use_decimal_percent = strtobool(use_decimal_percent)
            except ValueError:
                use_decimal_percent = False

        if use_decimal_percent:
            self.percent_factor = 0.01
        else:
            self.percent_factor = 1.0

        self.key_value_logs = sim_params.get(
            "key_value_logs", RuntimeConfig().misc.get("key_value_logs", False)
        )
        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        entities = list()
        params = model_params["params"]
        inits = model_params["inits"]
        self.num_models.setdefault(model, 0)

        for _ in range(num):

            eid = f"{model}-{self.num_models[model]}"
            self.models[eid] = MODELS[model](params, inits)
            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})

        return entities

    def step(self, time, inputs, max_advance=0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        LOG.debug(
            "At step %d: Received inputs: %s.",
            time,
            pprint.pformat(inputs),
        )

        self._set_default_inputs()

        # Set inputs from other simulators
        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():

                # Use time information from time generator
                if attr == "local_time":
                    self._set_attr_local_time(eid, src_ids)
                    continue

                attr_sum = self._aggregate_attr(src_ids)
                self._set_remaining_attrs(eid, attr, attr_sum)

        # Step the models
        for model in self.models.values():
            model.step()

        # Update time for the next step
        self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    def get_data(self, outputs):
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = dict()
        for eid, attrs in outputs.items():
            log_msg = {
                "id": f"{self.sid}_{eid}",
                "name": eid,
                "type": eid.split("-")[0],
            }

            for attr in attrs:
                value = self._get_remaining_attrs(eid, attr)
                data.setdefault(eid, dict())[attr] = value
                log_msg[attr] = value

            if self.key_value_logs:
                LOG.info(log_msg)

        if not self.key_value_logs:
            LOG.debug("Gathered outputs: %s.", pprint.pformat(data))

        return data

    def _set_default_inputs(self):
        VLOG.debug(
            "Setting step size %d and current time %s to all models.",
            self.step_size,
            self.now_dt,
        )
        for _, model in self.models.items():
            model.set_step_size(self.step_size)
            model.set_now_dt(self.now_dt)

    def _set_attr_local_time(self, eid: str, src_ids: Dict[str, Any]) -> bool:
        for val in src_ids.values():
            self.models[eid].set_now_dt(val)
            self.now_dt = datetime.strptime(val, GER).astimezone(timezone.utc)
            return True

        return False

    def _aggregate_attr(self, src_ids: Dict[str, Any]) -> float:
        """Aggregate inputs from different sources.

        If more inputs for one source exists, the average is calculated.

        """
        attr_sum = 0
        for val in src_ids.values():
            if val is None:
                continue
            if isinstance(val, (list, np.ndarray)):
                # This should only happen if palaestrAI is used
                val = val[0]
            attr_sum += val
        attr_sum /= len(src_ids)

        return float(attr_sum)

    def _set_percent_power(self, eid: str, attr_sum: float):
        attr_sum /= self.percent_factor
        attr_sum = self.models[eid].get_pn_min_kw() + attr_sum * (
            self.models[eid].get_pn_max_kw() - self.models[eid].get_pn_min_kw()
        )
        self.models[eid].set_p_kw(attr_sum)

    def _set_remaining_attrs(self, eid: str, attr: str, attr_sum: float):
        # Apply corrections
        if attr in ("p_set_mw", "p_th_set_mw", "q_set_mvar"):
            attr = attr.replace("m", "k")
            attr_sum *= 1e3

        # Set the inputs
        if attr == "set_percent":
            self._set_percent_power(eid, attr_sum)
        elif attr == "p_set_kw":
            self.models[eid].set_p_kw(attr_sum)
        elif attr == "q_set_kvar":
            self.models[eid].set_q_kvar(attr_sum)
        else:
            setattr(self.models[eid].inputs, attr, attr_sum)

    def _get_remaining_attrs(self, eid: str, attr: str) -> float:
        # Apply correction of the attr if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar"):
            true_attr = attr.replace("m", "k")
        else:
            true_attr = attr

        if true_attr == "p_kw":
            value = self.models[eid].get_p_kw()
        elif true_attr == "q_kvar":
            value = self.models[eid].get_q_kvar()
        else:
            value = getattr(self.models[eid].state, true_attr)

        # Apply correction of the value if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar"):
            value *= 1e-3

        return value


if __name__ == "__main__":
    set_and_init_logger(
        0, "pysimmods-logfile", "pysimmods-mosaik.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api.start_simulation(PysimmodsSimulator())



from enum import Enum
from typing import Optional

from .generator import Generator


class QControl(Enum):
    PRIORITIZE_P = 0
    PRIORITIZE_Q = 1


# P_DOMINATED = [QControl.P_SET, QControl.PQ_SET]
# Q_DOMINATED = [QControl.Q_SET, QControl.QP_SET]
# COS_PHI_DOMINATED = [QControl.COS_PHI_SET]


class QGenerator(Generator):
    """A generator subtype model with reactive power support.

    The set_percent and get_percent_in functions behave differently
    depending on the q_control attribute of the model.
    """

    def set_q_kvar(self, q_kvar: Optional[float]) -> None:
        if q_kvar is None:
            self.inputs.q_set_kvar = None
        else:
            self.inputs.q_set_kvar = q_kvar * self.config.gsign

        # q_min = self.get_qn_min_kvar()
        # q_max = self.get_qn_max_kvar()

        # if q_kvar * self.config.gsign > 0:
        #     # Consider capacitive and inductive
        #     q_kvar = abs(q_kvar)
        #     if q_kvar < q_min:
        #         self.inputs.q_set_kvar = 0
        #     else:
        #         self.inputs.q_set_kvar = min(q_max, q_kvar)
        # else:
        #     self.inputs.q_set_kvar = 0

    def get_q_kvar(self) -> float:
        return self.state.q_kvar * self.config.gsign

    # def get_percent_in(self):

    #     if self.config.q_control in P_DOMINATED:
    #         p_kw = self.inputs.p_set_kw
    #         if p_kw is None:
    #             return None
    #         if p_kw > 0:
    #             val = abs(p_kw)
    #             val_min = self.get_pn_min_kw()
    #             val_max = self.get_pn_max_kw()
    #         else:
    #             return 0
    #     if self.config.q_control in Q_DOMINATED:
    #         q_kvar = self.inputs.q_set_kvar
    #         if q_kvar is None:
    #             return None
    #         if q_kvar > 0:
    #             val = abs(q_kvar)
    #             val_min = self.get_qn_min_kvar()
    #             val_max = self.get_qn_max_kvar()
    #         else:
    #             return 0
    #     if self.config.q_control in COS_PHI_DOMINATED:
    #         val = self.inputs.cos_phi
    #         if val is None:
    #             return None
    #         return val if self.config.use_decimal_percent else val * 100

    #     decimal = (val - val_min) / (val_max - val_min)
    #     if decimal < 0:
    #         return 0
    #     decimal = max(0.0001, decimal)

    #     if self.config.use_decimal_percent:
    #         return decimal
    #     else:
    #         return decimal * 100

    # def set_percent(self, percentage):

    #     self.inputs.cos_phi = self.config.cos_phi
    #     if self.config.q_control == QControl.P_SET:
    #         if percentage <= 0:
    #             self.inputs.p_set_kw = 0
    #             return
    #         self._set_percent_p(percentage)

    #     elif self.config.q_control == QControl.Q_SET:
    #         if percentage <= 0:
    #             self.inputs.q_set_kvar = 0
    #             return
    #         self._set_percent_q(percentage)

    #     elif self.config.q_control == QControl.PQ_SET:
    #         self.inputs.cos_phi = None
    #         if percentage <= 0:
    #             self.inputs.p_set_kw = 0
    #             self.inputs.q_set_kvar = 0
    #             return
    #         self._set_percent_p(percentage)
    #         self.inputs.q_set_kvar = math.sqrt(
    #             self.config.s_max_kva**2 - self.inputs.p_set_kw**2
    #         )

    #     elif self.config.q_control == QControl.QP_SET:
    #         self.inputs.cos_phi = None
    #         if percentage <= 0:
    #             self.inputs.p_set_kw = 0
    #             self.inputs.q_set_kvar = 0
    #             return
    #         self._set_percent_q(percentage)
    #         p_set_kw = math.sqrt(
    #             self.config.s_max_kva**2 - self.inputs.q_set_kvar**2
    #         )
    #         if p_set_kw < self.get_pn_min_kw():
    #             self.inputs.p_set_kw = 0
    #         else:
    #             self.inputs.p_set_kw = min(self.get_pn_max_kw(), p_set_kw)

    #     elif self.config.q_control == QControl.COS_PHI_SET:
    #         if self.config.use_decimal_percent:
    #             decimal = max(min(percentage, 1.0), 0.0)
    #         else:
    #             # Internally, decimal percentage is used.
    #             decimal = max(min(percentage, 100.0), 0.0) / 100.0

    #         self.inputs.cos_phi = decimal

    def get_qn_min_kvar(self) -> float:
        return 0

    def get_qn_max_kvar(self) -> float:
        return self.config.s_max_kva

    # def _set_percent_p(self, percentage):

    #     if self.config.use_decimal_percent:
    #         decimal = max(min(abs(percentage), 1.0), 0.0)
    #     else:
    #         # Internally, decimal percentage is used.
    #         decimal = max(min(abs(percentage), 100.0), 0.0) / 100.0

    #     if decimal == 0:
    #         self.inputs.p_set_kw = 0
    #     if decimal <= 0.0001:
    #         # Map values lower than 0.01 % to minimum power
    #         decimal = 0

    #     p_max_kw = self.get_pn_max_kw()
    #     p_min_kw = self.get_pn_min_kw()
    #     self.inputs.p_set_kw = p_min_kw + decimal * (p_max_kw - p_min_kw)

    # def _set_percent_q(self, percentage):
    #     if percentage <= 0:
    #         self.inputs.q_set_kvar = 0
    #         return
    #     if self.config.use_decimal_percent:
    #         decimal = max(min(abs(percentage), 1.0), 0.0)
    #     else:
    #         # Internally, decimal percentage is used.
    #         decimal = max(min(abs(percentage), 100.0), 0.0) / 100.0

    #     if decimal == 0:
    #         self.inputs.q_set_kvar = 0

    #     q_min = self.get_qn_min_kvar()
    #     q_max = self.get_qn_max_kvar()
    #     self.inputs.q_set_kvar = q_min + decimal * (q_max - q_min)



"""This module contains the :class:`ForecastModel`."""
from datetime import datetime, timedelta
from typing import Optional

import numpy as np
import pandas as pd
from pysimmods.model.model import Model

# from pysimmods.other.flexibility.schedule import Schedule
from pysimmods.other.flexibility.schedule_model import ScheduleModel

from . import LOG

# from pysimmods.util.date_util import GER


class ForecastModel(ScheduleModel):
    """The forecast model for all pysimmods.

    This class extends the schedule model and allows to forecast the
    next steps.

    Parameters
    ----------
    model: :class:`.Model`
        The model that should be wrapped by the schedule model.
    unit: str, optional
        The unit to be used by the schedule. Default is `kw` for kilo
        watt. Other options are `w` and `mw`.
    prioritize_setpoint: bool, optional
        If set to True, this model will prefer setpoints rather than
        schedule values if setpoints are present. Default is False.
    forecast_horizon_hours: float, optional
        The number of hours the model should create a forecast for.
        If the model needs a weather forecast, this weather forecast
        needs to be large enough.

    Attributes
    ----------
    model: :class:`.Model`
        A reference to the model.
    now_dt: :class:`.datetime.datetime`
        The current local time.
    step_size: int
        The step size of the model.
    forecasts: :class:`pandas.DataFrame`
        A dictionary containing forecasts for the inputs of the
        underlying model.
    flexibilities: dict
        A dictionary containing the current flexibilities of the
        underlying model.
    schedule: :class:`.Schedule`
        Contains the current schedule of the model.

    """

    def __init__(
        self,
        model: Model,
        unit="kw",
        prioritize_setpoint: bool = False,
        step_size: Optional[int] = None,
        now_dt: Optional[datetime] = None,
        forecast_horizon_hours: float = 1,
    ):
        super().__init__(model, unit, prioritize_setpoint, step_size, now_dt)

        self._forecasts: Optional[pd.DataFrame] = None
        self._horizon_hours = forecast_horizon_hours

    def step(self):
        """Perform a simulation step of the underlying model.

        Also updates the internal state of the flexibility model.

        """
        super().step()

        self._check_schedule()

    def update_forecasts(self, forecasts):
        if self._forecasts is None:
            self._forecasts = forecasts
        else:
            for col in forecasts.columns:
                if col not in self._forecasts.columns:
                    self._forecasts[col] = np.nan
            for index, _ in forecasts.iterrows():
                if index not in self._forecasts.index:
                    break

            # Update existing entries
            self._forecasts.update(forecasts.loc[:index])
            # Add missing entries
            self._forecasts = pd.concat(
                [self._forecasts, forecasts.loc[index:]]
            )
            # Remove duplicates
            self._forecasts = self._forecasts[
                ~self._forecasts.index.duplicated()
            ]

    def _check_schedule(self):
        self.schedule.horizon_hours = self._horizon_hours

        if self.schedule.reschedule_required():
            self._create_default_schedule()
            self.schedule.prune()

    def _create_default_schedule(self):
        state_backup = self._model.get_state()

        now = self._now_dt + timedelta(seconds=self._step_size)
        periods = int(self._horizon_hours * 3_600 / self._step_size)

        for _ in range(periods):
            self._prepare_step(now)
            self._perform_step(now)

            now += timedelta(seconds=self._step_size)

        self._model.set_state(state_backup)

    def _prepare_step(self, now: datetime):
        if not self.schedule.has_index(now):
            self.schedule.update_row(now, np.nan, np.nan, np.nan, np.nan)

        default_p = self._model.get_default_p_set(now.hour)
        default_q = self._model.get_default_q_set(now.hour)

        if self.schedule.get(now, self._psetname) is None:
            self.schedule.update_entry(now, self._psetname, default_p)
        if self.schedule.get(now, self._qsetname) is None:
            self.schedule.update_entry(now, self._qsetname, default_q)

    def _perform_step(self, now):
        try:
            self._calculate_step(
                now,
                self.schedule.get(now, self._psetname),
                self.schedule.get(now, self._qsetname),
            )

            self.schedule.update_entry(
                now,
                self._pname,
                self._model.get_p_kw() * self._unit_factor,
            )
            self.schedule.update_entry(
                now,
                self._qname,
                self._model.get_q_kvar() * self._unit_factor,
            )

        except KeyError:
            # Forecast is missing
            LOG.info(
                "No forecast provided at %s for model %s.",
                now,
                self._model,
            )
            self.schedule.update_row(now, np.nan, np.nan, np.nan, np.nan)

    def _calculate_step(self, index, p_set, q_set):
        if p_set is not None and ~np.isnan(p_set):
            self._model.set_p_kw(p_set * self._unit_factor)
        if q_set is not None and ~np.isnan(q_set):
            self._model.set_q_kvar(q_set * self._unit_factor)

        if self._forecasts is not None:
            for col in self._forecasts.columns:
                if hasattr(self._model.inputs, col):
                    setattr(
                        self._model.inputs,
                        col,
                        self._forecasts.loc[index, col],
                    )

        self._model.set_now_dt(index)
        self._model.set_step_size(self._step_size)
        self._model.step()



import copy
import itertools
from copy import deepcopy

import numpy as np
from pysimmods.generator.biogassim.config import BiogasConfig
from pysimmods.generator.biogassim.inputs import BiogasInputs
from pysimmods.generator.biogassim.state import BiogasState
from pysimmods.generator.chpcngsim import CHPCNG
from pysimmods.model.generator import Generator


class BiogasPlant(Generator):
    """A biogas plant simulation model.

    A biogas plant consists basically of four main components:
    gas production, gas storage, one or more generator units, and
    a control unit. Additionally, a heat storage can be attached.

    These components are more or less integrated in the main model
    defined in this file. The generator units are out-sourced to
    pysimmods.chpngsim. A heat storage will be integrated in the
    future.

    An example parameter configuration for two generators could
    look like this::

        params = {
            'gas_m3_per_day': 3030,
            'cap_gas_m3': 1530,
            'gas_fill_min_percent': 2.0,
            'gas_fill_max_percent': 98.0,
            'ch4_concentration_percent': 54.4,
            'num_chps': 2,
            'chp0': {
                'pn_stages_kw': [75.0, 150.0],
                'eta_stages_percent': [31.1, 32.4],
                'eta_th_stages_percent': [30.8, 31.2],
                'restarts_per_day': 6,
                'active_min_s': 3 * 3_600,
                'active_max_s_per_day': 0,
                'inactive_min_s': 2 * 3_600,
                'inactive_max_s_per_day': 0,
            },
            'chp1': {
                'pn_stages_kw': [75.0, 150.0],
                'eta_stages_percent': [33.0, 33.2],
                'eta_th_stages_percent': [30.0, 30.5],
                'restarts_per_day': 6,
                'active_min_s': 3 * 3_600,
                'active_max_s_per_day': 0,
                'inactive_min_s': 2 * 3_600,
                'inactive_max_s_per_day': 0,
            }
        }

        inits = {
            'gas_fill_percent': 50,
            'chp0': {
                'active_s': 0,
                'active_s_per_day': 0,
                'inactive_s': 0,
                'inactive_s_per_day': 0,
                'restarts': 0,
                'p_kw': 75.0,
            },
            'chp1': {
                'active_s': 0,
                'active_s_per_day': 0,
                'inactive_s': 0,
                'inactive_s_per_day': 0,
                'restarts': 0,
                'p_kw': 75.0,
            },
        }

    More example configurations can be found in the presets.py

    Attributes
    ----------
    config: :class:`.BiogasConfig`
        The configuration parameters of the biogas model.
    state: :class:`.BiogasState`
        The initialization parameters of the biogas model.
    inputs: :class:`.BiogasInputs`
        The input parameters of the biogas model.
    chps: List[:class:`.CHPCNG`]
        A list containing the CHP objects of this biogas model.

    """

    def __init__(self, params, inits):
        self.config = BiogasConfig(params)
        self.state = BiogasState(inits)
        self.inputs = BiogasInputs()

        self.chps = list()
        for idx in range(self.config.num_chps):
            params[f"chp{idx}"].setdefault(
                "ch4_concentration_percent",
                self.config.ch4_concentration_percent,
            )
            params[f"chp{idx}"].setdefault(
                "sign_convention", self.config.sign_convention
            )
            self.chps.append(CHPCNG(params[f"chp{idx}"], inits[f"chp{idx}"]))

    def step(self):
        """Perform a simulation step."""
        next_state = deepcopy(self.state)
        self._check_inputs(next_state)
        self._check_gas_priority(next_state)
        self._check_gas_storage(next_state)

        self._step_chps(next_state)

        self._update_references(next_state)

        self.state = next_state
        self.inputs.reset()

    def _check_inputs(self, nstate):
        if self.inputs.p_set_kw is None:
            if self.inputs.now_dt is not None:
                nstate.p_kw = self.config.default_p_schedule[
                    self.inputs.now_dt.hour
                ]
            else:
                nstate.p_kw = 0
        else:
            nstate.p_kw = self.inputs.p_set_kw

    def _check_gas_priority(self, next_state):
        """Reset critical flags of the gas storage if feasible."""
        if 40 < self.state.gas_fill_percent < 60:
            next_state.burn_gas = False
            next_state.pool_gas = False

        if (
            self.config.gas_fill_min_percent
            < self.state.gas_fill_percent
            < self.config.gas_fill_max_percent
        ):
            next_state.gas_critical = False

    def _check_gas_storage(self, next_state):
        """Update the production part of the gas storage."""
        # Let's produce natural gas
        gas_per_s = self.config.gas_m3_per_day / 86_400
        next_state.gas_prod_m3 = gas_per_s * self.inputs.step_size

        # Now fill the gas storage
        gas_level = self.state.gas_fill_percent * 0.01 * self.config.cap_gas_m3
        gas_level += next_state.gas_prod_m3
        next_state.gas_fill_percent = gas_level / self.config.cap_gas_m3 * 100

    def _step_chps(self, next_state):
        """Find a suitable combination to step the CHPs to reach
        target electrical power (next_state.p_kw).
        """

        # Generate all combinations
        p_combis = list(
            itertools.product(
                *[[0] + chp.config.pn_stages_kw for chp in self.chps]
            )
        )

        # Try every combination and drop infeasible ones
        feasibles = self._generate_feasible_combis(next_state, p_combis)

        if not feasibles:
            # TODO: add logger warning
            feasibles[p_combis[0]] = 0

        # Determine the best combination
        best_score = np.Infinity
        best_combi = None
        for combi, score in feasibles.items():
            if score <= best_score:
                best_combi = combi
                best_score = score

        # Perform the steps with the best combination
        for chp, p_set_kw in zip(self.chps, best_combi):
            self._set_chp_inputs(next_state)
            # TODO: Adapt available gas
            chp.inputs.p_set_kw = p_set_kw
            chp.step()

    def _generate_feasible_combis(self, next_state, p_combis):

        feasibles = dict()
        for combi in p_combis:
            states = list()
            for chp, p_set_kw in zip(self.chps, combi):
                self._set_chp_inputs(next_state)
                chp.inputs.p_set_kw = p_set_kw
                states.append(chp.step(pretend=True))

            p_kws = list()
            gas_cons = list()
            for state in states:
                p_kws.append(state.p_kw)
                gas_cons.append(state.gas_cons_m3)

            diff = abs(sum(p_kws) - next_state.p_kw)

            new_gas_level = (
                min(100.0, next_state.gas_fill_percent)
                * 0.01
                * self.config.cap_gas_m3
            )
            new_gas_level -= sum(gas_cons)
            new_gas_percent = new_gas_level / self.config.cap_gas_m3 * 100

            if new_gas_percent < self.config.gas_fill_min_percent:
                if new_gas_percent < next_state.gas_fill_percent:
                    continue
            if new_gas_percent > self.config.gas_fill_max_percent:
                if new_gas_percent > next_state.gas_fill_percent:
                    continue

            feasibles[combi] = diff

        return feasibles

    def _set_chp_inputs(self, next_state):
        # Update inputs of the CHPs
        gas_level = next_state.gas_fill_percent * 0.01 * self.config.cap_gas_m3
        for chp in self.chps:
            chp.set_step_size(self.inputs.step_size)
            chp.set_now_dt(self.inputs.now_dt)
            chp.inputs.gas_critical = (
                next_state.burn_gas or next_state.pool_gas
            )
            chp.inputs.gas_in_m3 = gas_level / len(self.chps)

    def _update_references(self, next_state):
        next_state.p_kw = sum([chp.state.p_kw for chp in self.chps])
        next_state.p_th_kw = sum([chp.state.p_th_kw for chp in self.chps])

        next_state.gas_cons_m3 = sum(
            [chp.state.gas_cons_m3 for chp in self.chps]
        )
        next_state.gas_fill_percent -= (
            next_state.gas_cons_m3 / self.config.cap_gas_m3 * 100
        )
        next_state.gas_fill_percent = min(
            100, max(0, next_state.gas_fill_percent)
        )

        if next_state.gas_fill_percent < self.config.gas_fill_min_percent:
            next_state.pool_gas = True
            next_state.gas_critical = True

        if next_state.gas_fill_percent > self.config.gas_fill_max_percent:
            next_state.burn_gas = True
            next_state.gas_critical = True

    def get_state(self):

        state = {"state": copy.deepcopy(self.state.__dict__)}
        for idx, chp in enumerate(self.chps):
            state[f"chp-{idx}"] = chp.get_state()
        return state

    def set_state(self, state):
        for attr, value in state["state"].items():
            setattr(self.state, attr, value)
        # self.state = state["state"]
        for idx, chp in enumerate(self.chps):
            chp.set_state(state[f"chp-{idx}"])



"""This module contains the input information of the heat demand
model.
"""


class HeatDemandInputs:
    """Captures the inputs of the heat demand model"""

    def __init__(self):
        self.step_size = None
        self.day_avg_t_air_deg_celsius = None



"""This module contains the base class for all model states."""
from datetime import datetime
from typing import Optional


class ModelState:
    """Base class for model states.

    Parameters
    ----------
    inits : dict
        A *dict* containing the state variables.

    Attributes
    ----------
    p_kw: float
        Current (or last) electrical active power P in [kW].
    q_kvar: float
        Current (or last) electrical reactive power Q in [kVAr].
    s_kva: float
        Current (or last) electrical apparent power S in [kVA].
    now_dt: datetime, optional
        Current (or last) date and time. Marks the time where the last
        step started.
    delta_s: int, optional
        Duration of the current (or last) step in [s].
    """

    def __init__(self, inits: dict) -> None:
        self.p_kw: float = inits.get("p_kw", 0)
        self.q_kvar: float = inits.get("q_kvar", 0)
        self.s_kva: float = inits.get("s_kva", 0)
        self.now_dt: Optional[datetime] = inits.get("now_dt", None)
        self.delta_s: Optional[int] = inits.get("delta_s", None)



from copy import copy

from pysimmods.generator.dieselsim.config import DieselGenConfig
from pysimmods.generator.dieselsim.inputs import DieselGenInputs
from pysimmods.generator.dieselsim.state import DieselGenState
from pysimmods.model.generator import Generator


class DieselGenerator(Generator):
    def __init__(self, params, inits):
        super().__init__(params, inits)
        self.config = DieselGenConfig(params)
        self.state = DieselGenState(inits)
        self.inputs = DieselGenInputs()

    def step(self):
        nstate = copy(self.state)

        self._check_inputs(nstate)

        self.state = nstate

    def _check_inputs(self, nstate):
        """Check constraints for active power."""
        if self.inputs.p_set_kw is None:
            if self.inputs.now_dt is not None:
                nstate.p_kw = self.config.default_p_schedule[
                    self.inputs.now_dt.hour
                ]
            else:
                nstate.p_kw = 0
        else:
            nstate.p_kw = self.inputs.p_set_kw

    # @property
    # def set_percent(self):
    #     p_kw = self.inputs.p_set_kw
    #     if p_kw is None:
    #         p_kw = self.state.p_kw

    #     p_kw = (p_kw - self.config._p_min_kw) / (
    #         self.config._p_max_kw - self.config._p_min_kw
    #     )
    #     return abs(p_kw) * 100

    # @set_percent.setter
    # def set_percent(self, value):
    #     value = max(min(abs(value), 100.0), 0.0)

    #     if value == 0:
    #         p_set_kw = 0
    #     else:
    #         p_range = self.config._p_max_kw - self.config._p_min_kw
    #         p_set_kw = self.config._p_min_kw + value / 100 * p_range

    #     self.inputs.p_set_kw = p_set_kw



"""This module contains a function :func:`tilted_surface` to calculate
irradiance on tilted surfaces.

"""
from math import cos, radians, sin, tan
import logging

LOG = logging.getLogger(__name__)


def tilted_surface(
    direct_radiation,
    diffuse_radiation,
    surface_tilt,
    surface_azimuth,
    sun_elevation,
    sun_azimuth,
    albedo=0.2,
    model="klucher",
):
    """Compute radiation on tilted surface.

    Parameters
    ----------
    direct_radiation : float
        Direct solar radiation on horizontal surface in Watt/m2.
    diffuse_radiation : float
        Diffuse solar radiation on horizontal surface in Watt/m2.
    surface_tilt : float
        Tilt of the surface in degree (between 0° and 90°).
    surface_azimuth : float
        Orientation of the surface in degree (North 0°, East 90°,
        West 270°).
    sun_elevation : float
        Elevation of the sun in degree.
    sun_azimuth : float
        Solar azimuth angle in degree (Noth 0°, East 90°, West 270°).
    albedo: float
        Ratio of reflected radiation from surrounding surface to
        incident radiation on it.
    model : str
        Model which is used to compute solar radiation on tilted
        surface. Default is an advanced Klucher model [#]_.

    Returns
    -------
    float
        Total solar radiation on tilted surface in Watt/m2.

    Raises
    ------
    ValueError
        If either *direct_radiation* or *diffuse_radiation* is < 0.

    .. [#] Behrens, Tanja. Bestimmung der spektralen Solarstrahlung am Erdboden
           aus Satellitendaten zur Bewertung des Leistugnsverhaltens von
           Dünnschicht-Solarzellen. Von der Fakultät für Mathematik
           und Naturwissenschaften der Universität Oldenburg angenommene
           Disseratation zur Erlangung des Grades und Titels einer Doktorin der
           Naturwissenschaften, 2011, S. 36 f.

    """
    if direct_radiation < 0:
        LOG.warning("Direct radiation must be greater than zero")
        direct_radiation = 0
    if diffuse_radiation < 0:
        LOG.warning("Diffuse radiation must be greater than zero")
        diffuse_radiation = 0
    global_radiation = direct_radiation + diffuse_radiation
    if global_radiation <= 0:
        return 0

    # calculate ratio of titled and horizontal beam irradiance
    sun_zenith = min(90, 90 - sun_elevation)

    cos_incident = max(
        0,
        cos(radians(surface_tilt)) * cos(radians(sun_zenith))
        + sin(radians(surface_tilt))
        * sin(radians(sun_zenith))
        * cos(radians(sun_azimuth - surface_azimuth)),
    )
    # dot product of surface normal and solar angle

    ratio = min(10, cos_incident / cos(radians(sun_zenith)))
    # avoid unfavorable angles of incident which especially appear in winter

    # calculate total radiation on tilted surface using Klucher's 1979 model

    if model == "klucher":
        F = 1 - (diffuse_radiation / global_radiation) ** 2
        # anisotropy parameter

        a = ratio + (1 - cos(radians(surface_tilt))) * albedo / 2

        b = -ratio + (
            (1 + F * sin(radians(surface_tilt / 2)) ** 3)
            * (1 + F * cos_incident**2 * sin(radians(sun_zenith)) ** 3)
            * (1 + cos(radians(surface_tilt)))
            / 2
        )

        radiation_tilt = a * global_radiation + b * diffuse_radiation

    return radiation_tilt



from pysimmods.other.heatdemandsim.util import hprofiles
from pysimmods.other.heatdemandsim.config import HeatDemandConfig
from pysimmods.other.heatdemandsim.inputs import HeatDemandInputs
from pysimmods.other.heatdemandsim.state import HeatDemandState
from pysimmods.other.heatdemandsim.heatdemand import HeatDemand
